#!/usr/bin/perl

#############################################################################
# DECLARATIONS
#############################################################################
use strict;
use warnings;
use File::Copy;
use File::Find;
use Getopt::Long;
$Getopt::Long::order = $PERMUTE;
$Getopt::Long::passthrough = 1;
$Getopt::Long::autoabbrev  = 0;
no warnings 'File::Find';

# declare global variables
use vars qw($VERSION
            $help
            $sourceDir
            $barrier
            $targetDir
            @fileTypes
            @fileList
            @progress
           );

# set version
$VERSION = '0.1';

# declare signal handlers
$SIG{HUP}  = \&stop;
$SIG{INT}  = \&stop;
$SIG{QUIT} = \&stop;
$SIG{TERM} = \&stop;

#############################################################################
# MAIN
#############################################################################

# check arguments and set defaults
checkArgs();

# copyFiles
exit(copyFiles());

#############################################################################
sub stop
#############################################################################
{
  # enable 'direct printing'
  $| = 1;

  print "\nReceived stop signal...";

  # disable 'direct printing'
  $| = 0;

  exit(1);
}

#############################################################################
sub checkArgs
#############################################################################
{
  my $files = undef;
  
  # show help if there are no arguments
  $help = 1 if(@ARGV == 0);
  
  GetOptions('h|help'     => \$help,                  # help mode
             's|source=s' => \$sourceDir,             # directory to read from
             't|target=s' => \$targetDir,             # directory to write to
             'a|amount=i' => \$barrier,               # amount of data to be copied (in MB)
             'f|files=s'  => \$files,                 # specify file types
             
             '<>'         => \&handleUnknownParameter # unknown/rest
            );

  # show usage if help is selected
  usage() if($help);            
            
  # check if directories are set
  unless(defined($sourceDir)) { print "ERROR No source directory set!\n"; exit(1); }
  unless(defined($targetDir)) { print "ERROR No target directory set!\n"; exit(1); }        
  
  # check for trailing slash in directories => remove if present
  chop($sourceDir) if(substr($sourceDir, -1, 1) eq '/');
  chop($targetDir) if(substr($targetDir, -1, 1) eq '/');

  # check directories
  checkDirs($sourceDir, $targetDir);
  
  # set amount of data
  unless(defined $barrier)
  {
    # no amount set => copy all
    $barrier = 0;
  }
  elsif($barrier < 0)
  {
    print "ERROR Amount of data must not be less than null!\n";
    exit(1);
  }
  else
  {
    # convert MB to bytes
    $barrier *= 1024 * 1024;
  }

  # set default file types
  unless($files)
  {
    @fileTypes = ('MP3', 'OGG');
  }
  else
  {
    @fileTypes = split(',', uc($files));
  }
  
  # enable direct output (for progress)
  $| = 1;
  
  # set "progess animation"
  @progress    = ('|', '/', '-', '\\');
}

#############################################################################
sub handleUnknownParameter
#############################################################################
{
  print "ERROR Unknown parameter '$_[0]'\n";
  exit(1);
}

#############################################################################
sub usage
#############################################################################
{
  print "\n";
  print "$0 version $VERSION\n";
  print "\n";
  print "Copy files of selected type until a specified barrier is reached.\n";
  print "Usage: $0 [-h] -s <source> -t <target> [-a <SIZE>] [-f <LIST>]\n\n";
  print "  -h,--help           : show help\n";
  print "  -s,--source <DIR>   : directory to read from (incl. sub dirs)\n";
  print "  -t,--target <DIR>   : directory to write to\n";
  print "  -a,--amount <SIZE>  : amount of data (in MiB) to be copied\n";
  print "                        omit to copy all files\n";
  print "  -f,--files <LIST>   : comma separated list to specify file types\n";
  print "                        optional, default value = 'MP3,OGG'\n";
  print "                        Please arrange types in descending order relative to\n";
  print "                        its frequency of occurence.\n";
  print "\n";
  print "Example: $0 -s D:\\Music -t K:\\Phone -a 1024 -f mp3,ogg,wav\n";
  print "\n";
  print "  This copies at most 1024 MiB (1 GiB) of randomly selected mp3, ogg or wav\n";
  print "  files from D:\\Music to K:\\Phone. Files are overwritten without promt and\n";
  print "  the directory hierarchy gets lost/flattened!\n";
  print "\n";
  exit(1);
}

#############################################################################
sub checkDirs
#############################################################################
{
  foreach(@_)
  {
    unless(-d $_)
    {
      print "ERROR Directory '$_' doesn't exist!\n";
      exit(1);
    }
  }
}

#############################################################################
sub copyFiles
#############################################################################
{
  my $error   = 0;
  my $bytes   = 0;
  my $elems   = 0;
  my $random  = 0;
  my $copied  = 0;
    
  # init randomizer with PID and current time
  srand($$ ^ time());
  
  # create a list of all files
  generateFileList();
  
  # files found?
  unless($elems = scalar(@fileList))
  {
    print "No files found!\n";
    return 1;
  }
  
  print "Found $elems files in $sourceDir\n";
  
  while(1)
  {
    # generate random index
    $random = int(rand($elems));
  
    if(($bytes += -s $fileList[$random]) <= $barrier || $barrier == 0)
    {
      # copy file to target
      if(copy($fileList[$random], $targetDir))
      {
        # increase copied
        $copied++;
        # show progress
        showProgress($bytes, $copied);
        # delete current elem
        splice(@fileList, $random, 1);
        # decrease elems and stop if nothing is left
        unless(--$elems)
        {
          # show progress (100%)
          showProgress($barrier, $copied);
          print "\nNo files left in source directory!\n";
          last;
        }
      }
      else
      {
        print "\nERROR Failed to copy $fileList[$random] to $targetDir\nERROR: $!\n";
        $error = 1;
        last;
      }
    }
    elsif($barrier > 0)
    {
      # decrease bytes copied
      $bytes -= -s $fileList[$random];
      # show progress (100%)
      showProgress($barrier, $copied);
      # abort
      print "\nBarrier reached!\n";
      last;
    }
  }
  
  print "Copied $copied files (" . int($bytes/1024/1024) . " MB) to $targetDir\n";
  
  return $error;
}

#############################################################################
sub generateFileList
#############################################################################
{
  find(\&wanted, $sourceDir);
}

#############################################################################
sub wanted
#############################################################################
{
  my $file = $_;
  my $name = $File::Find::name;
  my $ext  = uc(($file =~ m/([^.]+)$/)[0]);

  foreach(@fileTypes)
  {
    if($ext eq $_)
    {
      push(@fileList, $name);
      return;
    }
  }
}

#############################################################################
sub showProgress
#############################################################################
{
  if($barrier)
  {
    my $hlp = $_[0] / $barrier;
 
    printf("\r%s%-69s [ %3d%% ]", $progress[$_[1]%3], '#' x ($hlp * 69), $hlp * 100);
  }
  else
  {
    printf("\r%s copying in progress", $progress[$_[1]%3]);
  }
}
